import * as dotenv from 'dotenv';
import { Request, Response } from 'express';
import { Options } from 'express-rate-limit';
import * as fs from 'fs';
import * as Joi from 'joi';
import { EnvConfig } from './env-config.interface';

export class ConfigService {
  private readonly envConfig: EnvConfig;

  constructor(filePath: string) {
    let config = {};
    if (fs.existsSync(filePath)) {
      config = dotenv.parse(fs.readFileSync(filePath));
    }
    this.envConfig = this.validateInput({ ...config, ...process.env });
  }

  /**
   * Ensures all needed variables are set, and returns the validated JavaScript object
   * including the applied default values.
   */
  private validateInput(envConfig: EnvConfig): EnvConfig {
    const envVarsSchema: Joi.ObjectSchema = Joi.object({
      API_KEY: Joi.string().required(),
      APP_LOG_DIR: Joi.string().required(),
      APP_LOG_FILES_PREFIX: Joi.string().required(),
      APP_LOG_MAX_LEVEL: Joi.string().required(),
      BODY_PARSER_LIMIT: Joi.string().default('50mb'),
      DOCUMENTS_DIR: Joi.string().required(),
      NODE_ENV: Joi.string().valid('dev', 'qualif', 'demo', 'prod', 'metro', 'workflow').default('dev'),
      RATE_LIMIT_ENABLED: Joi.boolean().default(true),
      RATE_LIMIT_HEADERS: Joi.boolean().default(false),
      RATE_LIMIT_MAX: Joi.number().default(100),
      RATE_LIMIT_WINDOW: Joi.number().default(300000),
    });

    const options = {
      allowUnknown: true,
      convert: true,
    };

    const { error, value: validatedEnvConfig } = envVarsSchema.validate(envConfig, options);
    if (error) {
      throw new Error(`Config validation error: ${error.message}`);
    }
    return validatedEnvConfig;
  }

  public readonly get: <T = string>(key: string) => T = <T = string>(key: string): T => this.envConfig[key] as unknown as T;

  public get apiKey(): string {
    return String(this.envConfig.API_KEY);
  }

  public get appLogDir(): string {
    return String(this.envConfig.APP_LOG_DIR);
  }

  public get appLogFilesPrefix(): string {
    return String(this.envConfig.APP_LOG_FILES_PREFIX);
  }

  public get appLogMaxLevel(): string {
    return String(this.envConfig.APP_LOG_MAX_LEVEL);
  }

  public get bodyParserLimit(): string {
    return String(this.envConfig.BODY_PARSER_LIMIT);
  }

  public get documentsDir(): string {
    return String(this.envConfig.DOCUMENTS_DIR);
  }

  public get nodeEnv(): string {
    return String(this.envConfig.NODE_ENV);
  }

  public get rateLimit(): Options {
    return {
      windowMs: Number(this.envConfig.RATE_LIMIT_WINDOW),
      max: Number(this.envConfig.RATE_LIMIT_MAX),
      standardHeaders: Boolean(this.envConfig.RATE_LIMIT_HEADERS),
      skip: (req: Request, res: Response) => Boolean(this.envConfig.RATE_LIMIT_ENABLED) !== true,
    } as Options;
  }
}
