import { ConsoleLogger, Injectable, LogLevel, Scope } from '@nestjs/common';
import * as winston from 'winston';
import * as DailyRotateFile from 'winston-daily-rotate-file';

@Injectable({ scope: Scope.TRANSIENT })
export class LoggerService extends ConsoleLogger {
  appLogMaxLevel: LogLevel;
  appLogMaxLevelDefault: LogLevel = 'log';
  levelsNotLoggedInConsole: LogLevel[]; // Don't want to display debug/verbose in console

  constructor(context: string, appLogDir: string, appLogFilesPrefix: string, appLogMaxLevel: LogLevel) {
    super(context);

    if (this.nestLevels.indexOf(appLogMaxLevel) < 0) {
      super.warn(
        `Unknown NestJS log level : '${this.appLogMaxLevel}'. Must be one of : 'error', 'warn', 'log', 'debug', 'verbose'. Setting '${this.appLogMaxLevelDefault}' by default.`,
      );
      this.appLogMaxLevel = this.appLogMaxLevelDefault;
    } else {
      this.appLogMaxLevel = appLogMaxLevel;
    }

    this.levelsNotLoggedInConsole = this.nestLevels.slice(this.nestLevels.indexOf(this.appLogMaxLevel) + 1);

    this.createWinstonLoggers(appLogDir, appLogFilesPrefix);
  }

  // *********************************************************************************************
  // * WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON W *
  // * WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON WINSTON W *
  // *********************************************************************************************

  winstonLoggers: winston.Logger[] = [];

  private createWinstonLoggers(appLogDir: string, appLogFilesPrefix: string): void {
    const dailyOptions: DailyRotateFile.DailyRotateFileTransportOptions = {
      zippedArchive: false,
      maxSize: '20m',
      maxFiles: '14d',
    };

    // *********************************************************************************************
    // * Formatters

    const readableFormat: winston.Logform.Format = winston.format.combine(
      winston.format.errors({ stack: true }), // Cf. https://stackoverflow.com/a/58475687
      winston.format.timestamp({ format: 'DD/MM/YYYY HH:mm:ss' }),
      winston.format.printf(
        ({ level, message, timestamp, stack, ...meta }) =>
          `${timestamp}     ${level.toUpperCase()} [${meta[Symbol.for('splat')]}] ${level == 'error' ? stack : message}`,
      ),
    );

    // *********************************************************************************************
    // * Loggers

    (
      [
        { level: 'error', filePrefix: appLogFilesPrefix + '-error-readable-' },
        { level: this.appLogMaxLevel, filePrefix: appLogFilesPrefix + '-all-readable-' },
      ] as { level: LogLevel; filePrefix: string }[]
    ).forEach((logger) => {
      try {
        this.winstonLoggers.push(
          winston.createLogger({
            level: logger.level,
            format: readableFormat,
            transports: [
              new DailyRotateFile({
                filename: appLogDir + `/${logger.filePrefix}%DATE%.log`,
                ...dailyOptions,
              } as DailyRotateFile.DailyRotateFileTransportOptions),
            ],
          }),
        );
      } catch (error) {
        console.warn(`Unable to create winswton logger '${logger.level} / ${logger.filePrefix}' : ${error}`);
      }
    });
  }

  // *********************************************************************************************
  // * LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVEL *
  // * LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVELS LEVEL *
  // *********************************************************************************************

  private nestLevels: LogLevel[] = ['error', 'warn', 'log', 'debug', 'verbose'];

  private fromNestToWinston(level: string): string {
    return level === 'log' ? 'info' : level;
  }

  error(message: any, context?: string): void {
    this.displayMsg(message, 'error', context);
  }

  warn(message: any, context?: string): void {
    this.displayMsg(message, 'warn', context);
  }

  log(message: any, context?: string): void {
    this.displayMsg(message, 'log', context);
  }

  debug(message: any, context?: string): void {
    this.displayMsg(message, 'debug', context);
  }

  verbose(message: any, context?: string): void {
    this.displayMsg(message, 'verbose', context);
  }

  // *********************************************************************************************
  // * MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG M *
  // * MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG MSG M *
  // *********************************************************************************************

  private displayMsg(message: any, level: LogLevel, context?: string): void {
    // NestJS logger
    if (!this.levelsNotLoggedInConsole.includes(level)) {
      super[level].apply(this, [message, context]);
    }

    // Winston log
    this.winstonLoggers.forEach((logger: winston.Logger) => {
      try {
        logger.log(this.fromNestToWinston(level), message, context);
      } catch (error) {
        console.warn(`Unable to add log ${message} into winston logger : ${error}`);
      }
    });
  }
}
