import { INestApplication, LogLevel, Logger, ValidationPipe } from '@nestjs/common';
import { NestFactory } from '@nestjs/core';
import * as express from 'express';
import * as RateLimit from 'express-rate-limit';
import * as helmet from 'helmet';
import * as https from 'https';
import { AppModule } from './app.module';
import { ConfigService } from './config/config.service';
import { LoggerService } from './logger/logger.service';
import { SwaggerDocumentProvider } from './swagger-document.provider';

async function bootstrap() {
  const app: INestApplication = await NestFactory.create(AppModule);
  const config: ConfigService = app.get(ConfigService);

  app.useGlobalPipes(new ValidationPipe());
  app.use(helmet.default());
  app.use(
    // @ts-ignore: known issue
    RateLimit({ ...config.rateLimit }),
  );

  app.use(express.json({ limit: config.bodyParserLimit }));
  app.use(express.urlencoded({ limit: config.bodyParserLimit, extended: true }));

  // Logger
  app.useLogger(new LoggerService('AppModule', config.appLogDir, config.appLogFilesPrefix, <LogLevel>config.appLogMaxLevel));

  // Skip SSL certificate verification
  https.globalAgent.options.rejectUnauthorized = false;

  // launch swagger only when environment is not prod
  if (config.nodeEnv === 'dev') {
    SwaggerDocumentProvider.use(app);
  }
  app.enableCors();

  // Handle unhandled rejections - Prevent node process to exit systematically
  process.on('unhandledRejection', (reason: any, promise) => {
    Logger.error(`Unhandled Promise rejection : ${reason}`, 'main.ts');
    console.error(reason?.stack);

    const unstableServerState: boolean = false; // Do we identify that the server state became unstable ?
    if (unstableServerState) {
      app.close().then(() => {
        // Better to throw an error rather than to call process.exit(…),
        // so NodeJS process can finish asynchronous processes.
        throw Error(`Unhandled Promise rejection : ${reason}`);
      });
    } else {
      console.log('NestJS server stays running.');
    }
  });

  await app.listen(9100);
}

bootstrap();
