import { Injectable, Logger, NotFoundException } from '@nestjs/common';
import { ensureDir, existsSync, move, readFile, readdir, removeSync, stat, unlink, writeFile } from 'fs-extra';
import { join } from 'path';
import { ConfigService } from '../../config/config.service';

@Injectable()
export class FileService {
  private readonly logger: Logger = new Logger(FileService.name);

  constructor(private readonly configService: ConfigService) {}

  // *********************************************************************************************
  // * READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ *
  // * READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ READ *
  // *********************************************************************************************

  public async readFile(filepath: string): Promise<Buffer> {
    const documentFilePath: string = this.configService.documentsDir + '/' + filepath;

    this.logger.debug(`Reading file ${documentFilePath}`);

    if (existsSync(documentFilePath)) {
      return readFile(documentFilePath);
    }

    this.logger.debug(`No such file or directory '${filepath}'`);
    throw new NotFoundException(`No such file or directory '${filepath}'`);
  }

  // *********************************************************************************************
  // * WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE *
  // * WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE WRITE *
  // *********************************************************************************************

  public async writeFileToDirectory(path: string, fullName: string, buffer: Buffer): Promise<void> {
    const documentFullPath: string = this.configService.documentsDir + '/' + path;
    const documentFullName: string = this.configService.documentsDir + '/' + fullName;

    this.logger.debug(`Writing-file-to-directory ${buffer?.length} chars into file ${documentFullName} (in path ${documentFullPath})`);

    await ensureDir(documentFullPath);
    return writeFile(documentFullName, buffer);
  }

  public async moveDoc(srcPath: string, destPath: string): Promise<void> {
    const documentSrcPath: string = this.configService.documentsDir + '/' + srcPath;
    const documentDescPath: string = this.configService.documentsDir + '/' + destPath;

    this.logger.debug(`Moving file ${documentSrcPath} to ${documentDescPath}`);

    return move(documentSrcPath, documentDescPath);
  }

  // *********************************************************************************************
  // * REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOV *
  // * REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOVE REMOV *
  // *********************************************************************************************

  public remove(path: string): void {
    const fullpath: string = join(this.configService.documentsDir, path);

    this.logger.debug(`Removing path ${fullpath}`);

    if (existsSync(fullpath)) {
      return removeSync(fullpath);
    }

    this.logger.debug(`No such file or directory ${path}`);
    throw new NotFoundException(`No such file or directory '${path}'`);
  }

  // *********************************************************************************************
  // * VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS V *
  // * VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS VARIOUS V *
  // *********************************************************************************************

  public async fileRotate(folder: string, durationDays: number): Promise<void> {
    const fullFolder: string = join(this.configService.documentsDir, folder);

    this.logger.debug(`File rotating folder ${fullFolder}) on duration ${durationDays}`);

    const now: Date = new Date();
    const dateToRemove: number = now.setDate(now.getDate() - durationDays);

    readdir(fullFolder, (err, files) => {
      if (err) {
        this.logger.error(`An error occurs when reading content of fullFolder ${fullFolder} : ${JSON.stringify(err)} `);
      }
      files.forEach((file) => {
        let path: string = join(fullFolder, file);
        stat(path, (err, stats) => {
          if (err) {
            this.logger.error(`An error occurs when reading stats of file ${path} : ${JSON.stringify(err)}`);
          }
          let fileInfo = { fileDate: stats.mtime, fileName: file };
          if (fileInfo.fileDate < new Date(dateToRemove)) {
            path = join(fullFolder, fileInfo.fileName);
            unlink(path);
          }
        });
      });
    });
  }
}
